@@ -19,6 +19,7 @@
 
 #include "AckGroupingTrackerEnabled.h"
 
+#include <climits>
 #include <mutex>
 
 #include "ClientConnection.h"
@@ -26,31 +27,21 @@
 #include "Commands.h"
 #include "ExecutorService.h"
 #include "HandlerBase.h"
-#include "LogUtils.h"
+#include "MessageIdUtil.h"
 
 namespace pulsar {
 
-DECLARE_LOG_OBJECT();
-
-AckGroupingTrackerEnabled::AckGroupingTrackerEnabled(ClientImplPtr clientPtr,
-                                                     const HandlerBasePtr& handlerPtr, uint64_t consumerId,
-                                                     long ackGroupingTimeMs, long ackGroupingMaxSize)
-    : AckGroupingTracker(),
-      isClosed_(false),
-      handlerWeakPtr_(handlerPtr),
-      consumerId_(consumerId),
-      nextCumulativeAckMsgId_(MessageId::earliest()),
-      requireCumulativeAck_(false),
-      mutexCumulativeAckMsgId_(),
-      pendingIndividualAcks_(),
-      rmutexPendingIndAcks_(),
-      ackGroupingTimeMs_(ackGroupingTimeMs),
-      ackGroupingMaxSize_(ackGroupingMaxSize),
-      executor_(clientPtr->getIOExecutorProvider()->get()),
-      timer_(),
-      mutexTimer_() {
-    LOG_DEBUG("ACK grouping is enabled, grouping time " << ackGroupingTimeMs << "ms, grouping max size "
-                                                        << ackGroupingMaxSize);
+// Define a customized compare logic whose difference with the default compare logic of MessageId is:
+// When two MessageId objects are in the same entry, if only one of them is a message in the batch, treat
+// it as a smaller one.
+static int compare(const MessageId& lhs, const MessageId& rhs) {
+    int result = compareLedgerAndEntryId(lhs, rhs);
+    if (result != 0) {
+        return result;
+    } else {
+        return internal::compare(lhs.batchIndex() < 0 ? INT_MAX : lhs.batchIndex(),
+                                 rhs.batchIndex() < 0 ? INT_MAX : rhs.batchIndex());
+    }
 }
 
 void AckGroupingTrackerEnabled::start() { this->scheduleTimer(); }
@@ -59,7 +50,7 @@ bool AckGroupingTrackerEnabled::isDuplicate(const MessageId& msgId) {
     {
         // Check if the message ID is already ACKed by a previous (or pending) cumulative request.
         std::lock_guard<std::mutex> lock(this->mutexCumulativeAckMsgId_);
-        if (msgId <= this->nextCumulativeAckMsgId_) {
+        if (compare(msgId, this->nextCumulativeAckMsgId_) <= 0) {
             return true;
         }
     }
@@ -69,29 +60,55 @@ bool AckGroupingTrackerEnabled::isDuplicate(const MessageId& msgId) {
     return this->pendingIndividualAcks_.count(msgId) > 0;
 }
 
-void AckGroupingTrackerEnabled::addAcknowledge(const MessageId& msgId) {
+void AckGroupingTrackerEnabled::addAcknowledge(const MessageId& msgId, ResultCallback callback) {
     std::lock_guard<std::recursive_mutex> lock(this->rmutexPendingIndAcks_);
     this->pendingIndividualAcks_.insert(msgId);
+    if (waitResponse_) {
+        this->pendingIndividualCallbacks_.emplace_back(callback);
+    } else if (callback) {
+        callback(ResultOk);
+    }
     if (this->ackGroupingMaxSize_ > 0 && this->pendingIndividualAcks_.size() >= this->ackGroupingMaxSize_) {
         this->flush();
     }
 }
 
-void AckGroupingTrackerEnabled::addAcknowledgeList(const MessageIdList& msgIds) {
+void AckGroupingTrackerEnabled::addAcknowledgeList(const MessageIdList& msgIds, ResultCallback callback) {
     std::lock_guard<std::recursive_mutex> lock(this->rmutexPendingIndAcks_);
     for (const auto& msgId : msgIds) {
         this->pendingIndividualAcks_.emplace(msgId);
     }
+    if (waitResponse_) {
+        this->pendingIndividualCallbacks_.emplace_back(callback);
+    } else if (callback) {
+        callback(ResultOk);
+    }
     if (this->ackGroupingMaxSize_ > 0 && this->pendingIndividualAcks_.size() >= this->ackGroupingMaxSize_) {
         this->flush();
     }
 }
 
-void AckGroupingTrackerEnabled::addAcknowledgeCumulative(const MessageId& msgId) {
-    std::lock_guard<std::mutex> lock(this->mutexCumulativeAckMsgId_);
-    if (msgId > this->nextCumulativeAckMsgId_) {
+void AckGroupingTrackerEnabled::addAcknowledgeCumulative(const MessageId& msgId, ResultCallback callback) {
+    std::unique_lock<std::mutex> lock(this->mutexCumulativeAckMsgId_);
+    if (compare(msgId, this->nextCumulativeAckMsgId_) > 0) {
         this->nextCumulativeAckMsgId_ = msgId;
         this->requireCumulativeAck_ = true;
+        // Trigger the previous pending callback
+        if (latestCumulativeCallback_) {
+            latestCumulativeCallback_(ResultOk);
+        }
+        if (waitResponse_) {
+            // Move the callback to latestCumulativeCallback_ so that it will be triggered when receiving the
+            // AckResponse or being replaced by a newer MessageId
+            latestCumulativeCallback_ = callback;
+            callback = nullptr;
+        } else {
+            latestCumulativeCallback_ = nullptr;
+        }
+    }
+    lock.unlock();
+    if (callback) {
+        callback(ResultOk);
     }
 }
 
@@ -106,27 +123,13 @@ void AckGroupingTrackerEnabled::close() {
 }
 
 void AckGroupingTrackerEnabled::flush() {
-    auto handler = handlerWeakPtr_.lock();
-    if (!handler) {
-        LOG_DEBUG("Reference to the HandlerBase is not valid.");
-        return;
-    }
-    auto cnx = handler->getCnx().lock();
-    if (cnx == nullptr) {
-        LOG_DEBUG("Connection is not ready, grouping ACK failed.");
-        return;
-    }
-
     // Send ACK for cumulative ACK requests.
     {
         std::lock_guard<std::mutex> lock(this->mutexCumulativeAckMsgId_);
         if (this->requireCumulativeAck_) {
-            if (!this->doImmediateAck(cnx, this->consumerId_, this->nextCumulativeAckMsgId_,
-                                      CommandAck_AckType_Cumulative)) {
-                // Failed to send ACK.
-                LOG_WARN("Failed to send cumulative ACK.");
-                return;
-            }
+            this->doImmediateAck(this->nextCumulativeAckMsgId_, this->latestCumulativeCallback_,
+                                 CommandAck_AckType_Cumulative);
+            this->latestCumulativeCallback_ = nullptr;
             this->requireCumulativeAck_ = false;
         }
     }
@@ -134,7 +137,14 @@ void AckGroupingTrackerEnabled::flush() {
     // Send ACK for individual ACK requests.
     std::lock_guard<std::recursive_mutex> lock(this->rmutexPendingIndAcks_);
     if (!this->pendingIndividualAcks_.empty()) {
-        this->doImmediateAck(cnx, consumerId_, this->pendingIndividualAcks_);
+        std::vector<ResultCallback> callbacks;
+        callbacks.swap(this->pendingIndividualCallbacks_);
+        auto callback = [callbacks](Result result) {
+            for (auto&& callback : callbacks) {
+                callback(result);
+            }
+        };
+        this->doImmediateAck(this->pendingIndividualAcks_, callback);
         this->pendingIndividualAcks_.clear();
     }
 }
@@ -144,6 +154,7 @@ void AckGroupingTrackerEnabled::flushAndClean() {
     {
         std::lock_guard<std::mutex> lock(this->mutexCumulativeAckMsgId_);
         this->nextCumulativeAckMsgId_ = MessageId::earliest();
+        this->latestCumulativeCallback_ = nullptr;
         this->requireCumulativeAck_ = false;
     }
     std::lock_guard<std::recursive_mutex> lock(this->rmutexPendingIndAcks_);
