@@ -19,13 +19,16 @@
 #include "ClientImpl.h"
 
 #include <pulsar/ClientConfiguration.h>
+#include <pulsar/Version.h>
 
 #include <random>
+#include <sstream>
 
 #include "BinaryProtoLookupService.h"
 #include "ClientConfigurationImpl.h"
 #include "Commands.h"
 #include "ConsumerImpl.h"
+#include "ConsumerInterceptors.h"
 #include "ExecutorService.h"
 #include "HTTPLookupService.h"
 #include "LogUtils.h"
@@ -33,15 +36,13 @@
 #include "PartitionedProducerImpl.h"
 #include "PatternMultiTopicsConsumerImpl.h"
 #include "ProducerImpl.h"
+#include "ProducerInterceptors.h"
 #include "ReaderImpl.h"
 #include "RetryableLookupService.h"
+#include "TableViewImpl.h"
 #include "TimeUtils.h"
 #include "TopicName.h"
 
-#ifdef USE_LOG4CXX
-#include "Log4CxxLogger.h"
-#endif
-
 #ifdef PULSAR_USE_BOOST_REGEX
 #include <boost/regex.hpp>
 #define PULSAR_REGEX_NAMESPACE boost
@@ -74,8 +75,7 @@ typedef std::unique_lock<std::mutex> Lock;
 
 typedef std::vector<std::string> StringList;
 
-ClientImpl::ClientImpl(const std::string& serviceUrl, const ClientConfiguration& clientConfiguration,
-                       bool poolConnections)
+ClientImpl::ClientImpl(const std::string& serviceUrl, const ClientConfiguration& clientConfiguration)
     : mutex_(),
       state_(Open),
       serviceNameResolver_(serviceUrl),
@@ -86,25 +86,15 @@ ClientImpl::ClientImpl(const std::string& serviceUrl, const ClientConfiguration&
           std::make_shared<ExecutorServiceProvider>(clientConfiguration_.getMessageListenerThreads())),
       partitionListenerExecutorProvider_(
           std::make_shared<ExecutorServiceProvider>(clientConfiguration_.getMessageListenerThreads())),
-      pool_(clientConfiguration_, ioExecutorProvider_, clientConfiguration_.getAuthPtr(), poolConnections),
+      pool_(clientConfiguration_, ioExecutorProvider_, clientConfiguration_.getAuthPtr(),
+            ClientImpl::getClientVersion(clientConfiguration)),
       producerIdGenerator_(0),
       consumerIdGenerator_(0),
-      requestIdGenerator_(0),
       closingError(ResultOk) {
     std::unique_ptr<LoggerFactory> loggerFactory = clientConfiguration_.impl_->takeLogger();
     if (!loggerFactory) {
-#ifdef USE_LOG4CXX
-        if (!clientConfiguration_.getLogConfFilePath().empty()) {
-            // A log4cxx log file was passed through deprecated parameter. Use that to configure Log4CXX
-            loggerFactory = Log4CxxLoggerFactory::create(clientConfiguration_.getLogConfFilePath());
-        } else {
-            // Use default simple console logger
-            loggerFactory.reset(new ConsoleLoggerFactory);
-        }
-#else
         // Use default simple console logger
         loggerFactory.reset(new ConsoleLoggerFactory);
-#endif
     }
     LogUtils::setLoggerFactory(std::move(loggerFactory));
 
@@ -116,9 +106,8 @@ ClientImpl::ClientImpl(const std::string& serviceUrl, const ClientConfiguration&
             std::cref(clientConfiguration_.getAuthPtr()));
     } else {
         LOG_DEBUG("Using Binary Lookup");
-        underlyingLookupServicePtr =
-            std::make_shared<BinaryProtoLookupService>(std::ref(serviceNameResolver_), std::ref(pool_),
-                                                       std::cref(clientConfiguration_.getListenerName()));
+        underlyingLookupServicePtr = std::make_shared<BinaryProtoLookupService>(
+            std::ref(serviceNameResolver_), std::ref(pool_), std::cref(clientConfiguration_));
     }
 
     lookupServicePtr_ = RetryableLookupService::create(
@@ -142,7 +131,7 @@ ExecutorServiceProviderPtr ClientImpl::getPartitionListenerExecutorProvider() {
 LookupServicePtr ClientImpl::getLookup() { return lookupServicePtr_; }
 
 void ClientImpl::createProducerAsync(const std::string& topic, ProducerConfiguration conf,
-                                     CreateProducerCallback callback) {
+                                     CreateProducerCallback callback, bool autoDownloadSchema) {
     if (conf.isChunkingEnabled() && conf.getBatchingEnabled()) {
         throw std::invalid_argument("Batching and chunking of messages can't be enabled together");
     }
@@ -159,9 +148,26 @@ void ClientImpl::createProducerAsync(const std::string& topic, ProducerConfigura
             return;
         }
     }
-    lookupServicePtr_->getPartitionMetadataAsync(topicName).addListener(
-        std::bind(&ClientImpl::handleCreateProducer, shared_from_this(), std::placeholders::_1,
-                  std::placeholders::_2, topicName, conf, callback));
+
+    if (autoDownloadSchema) {
+        auto self = shared_from_this();
+        lookupServicePtr_->getSchema(topicName).addListener(
+            [self, topicName, callback](Result res, SchemaInfo topicSchema) {
+                if (res != ResultOk) {
+                    callback(res, Producer());
+                    return;
+                }
+                ProducerConfiguration conf;
+                conf.setSchema(topicSchema);
+                self->lookupServicePtr_->getPartitionMetadataAsync(topicName).addListener(
+                    std::bind(&ClientImpl::handleCreateProducer, self, std::placeholders::_1,
+                              std::placeholders::_2, topicName, conf, callback));
+            });
+    } else {
+        lookupServicePtr_->getPartitionMetadataAsync(topicName).addListener(
+            std::bind(&ClientImpl::handleCreateProducer, shared_from_this(), std::placeholders::_1,
+                      std::placeholders::_2, topicName, conf, callback));
+    }
 }
 
 void ClientImpl::handleCreateProducer(const Result result, const LookupDataResultPtr partitionMetadata,
@@ -169,11 +175,20 @@ void ClientImpl::handleCreateProducer(const Result result, const LookupDataResul
                                       CreateProducerCallback callback) {
     if (!result) {
         ProducerImplBasePtr producer;
-        if (partitionMetadata->getPartitions() > 0) {
-            producer = std::make_shared<PartitionedProducerImpl>(shared_from_this(), topicName,
-                                                                 partitionMetadata->getPartitions(), conf);
-        } else {
-            producer = std::make_shared<ProducerImpl>(shared_from_this(), *topicName, conf);
+
+        auto interceptors = std::make_shared<ProducerInterceptors>(conf.getInterceptors());
+
+        try {
+            if (partitionMetadata->getPartitions() > 0) {
+                producer = std::make_shared<PartitionedProducerImpl>(
+                    shared_from_this(), topicName, partitionMetadata->getPartitions(), conf, interceptors);
+            } else {
+                producer = std::make_shared<ProducerImpl>(shared_from_this(), *topicName, conf, interceptors);
+            }
+        } catch (const std::runtime_error& e) {
+            LOG_ERROR("Failed to create producer: " << e.what());
+            callback(ResultConnectError, {});
+            return;
         }
         producer->getProducerCreatedFuture().addListener(
             std::bind(&ClientImpl::handleProducerCreated, shared_from_this(), std::placeholders::_1,
@@ -226,6 +241,33 @@ void ClientImpl::createReaderAsync(const std::string& topic, const MessageId& st
                   std::placeholders::_2, topicName, msgId, conf, callback));
 }
 
+void ClientImpl::createTableViewAsync(const std::string& topic, const TableViewConfiguration& conf,
+                                      TableViewCallback callback) {
+    TopicNamePtr topicName;
+    {
+        Lock lock(mutex_);
+        if (state_ != Open) {
+            lock.unlock();
+            callback(ResultAlreadyClosed, TableView());
+            return;
+        } else if (!(topicName = TopicName::get(topic))) {
+            lock.unlock();
+            callback(ResultInvalidTopicName, TableView());
+            return;
+        }
+    }
+
+    TableViewImplPtr tableViewPtr =
+        std::make_shared<TableViewImpl>(shared_from_this(), topicName->toString(), conf);
+    tableViewPtr->start().addListener([callback](Result result, TableViewImplPtr tableViewImplPtr) {
+        if (result == ResultOk) {
+            callback(result, TableView{tableViewImplPtr});
+        } else {
+            callback(result, {});
+        }
+    });
+}
+
 void ClientImpl::handleReaderMetadataLookup(const Result result, const LookupDataResultPtr partitionMetadata,
                                             TopicNamePtr topicName, MessageId startMessageId,
                                             ReaderConfiguration conf, ReaderCallback callback) {
@@ -236,15 +278,17 @@ void ClientImpl::handleReaderMetadataLookup(const Result result, const LookupDat
         return;
     }
 
-    if (partitionMetadata->getPartitions() > 0) {
-        LOG_ERROR("Topic reader cannot be created on a partitioned topic: " << topicName->toString());
-        callback(ResultOperationNotSupported, Reader());
+    ReaderImplPtr reader;
+    try {
+        reader.reset(new ReaderImpl(shared_from_this(), topicName->toString(),
+                                    partitionMetadata->getPartitions(), conf,
+                                    getListenerExecutorProvider()->get(), callback));
+    } catch (const std::runtime_error& e) {
+        LOG_ERROR("Failed to create reader: " << e.what());
+        callback(ResultConnectError, {});
         return;
     }
-
-    ReaderImplPtr reader = std::make_shared<ReaderImpl>(shared_from_this(), topicName->toString(), conf,
-                                                        getListenerExecutorProvider()->get(), callback);
-    ConsumerImplBasePtr consumer = reader->getConsumer().lock();
+    ConsumerImplBasePtr consumer = reader->getConsumer();
     auto self = shared_from_this();
     reader->start(startMessageId, [this, self](const ConsumerImplBaseWeakPtr& weakConsumerPtr) {
         auto consumer = weakConsumerPtr.lock();
@@ -280,28 +324,55 @@ void ClientImpl::subscribeWithRegexAsync(const std::string& regexPattern, const
         }
     }
 
-    NamespaceNamePtr nsName = topicNamePtr->getNamespaceName();
+    if (TopicName::containsDomain(regexPattern)) {
+        LOG_WARN("Ignore invalid domain: "
+                 << topicNamePtr->getDomain()
+                 << ", use the RegexSubscriptionMode parameter to set the topic type");
+    }
 
-    lookupServicePtr_->getTopicsOfNamespaceAsync(nsName).addListener(
-        std::bind(&ClientImpl::createPatternMultiTopicsConsumer, shared_from_this(), std::placeholders::_1,
-                  std::placeholders::_2, regexPattern, subscriptionName, conf, callback));
+    CommandGetTopicsOfNamespace_Mode mode;
+    auto regexSubscriptionMode = conf.getRegexSubscriptionMode();
+    switch (regexSubscriptionMode) {
+        case PersistentOnly:
+            mode = CommandGetTopicsOfNamespace_Mode_PERSISTENT;
+            break;
+        case NonPersistentOnly:
+            mode = CommandGetTopicsOfNamespace_Mode_NON_PERSISTENT;
+            break;
+        case AllTopics:
+            mode = CommandGetTopicsOfNamespace_Mode_ALL;
+            break;
+        default:
+            LOG_ERROR("RegexSubscriptionMode not valid: " << regexSubscriptionMode);
+            callback(ResultInvalidConfiguration, Consumer());
+            return;
+    }
+
+    lookupServicePtr_->getTopicsOfNamespaceAsync(topicNamePtr->getNamespaceName(), mode)
+        .addListener(std::bind(&ClientImpl::createPatternMultiTopicsConsumer, shared_from_this(),
+                               std::placeholders::_1, std::placeholders::_2, regexPattern, mode,
+                               subscriptionName, conf, callback));
 }
 
 void ClientImpl::createPatternMultiTopicsConsumer(const Result result, const NamespaceTopicsPtr topics,
                                                   const std::string& regexPattern,
+                                                  CommandGetTopicsOfNamespace_Mode mode,
                                                   const std::string& subscriptionName,
                                                   const ConsumerConfiguration& conf,
                                                   SubscribeCallback callback) {
     if (result == ResultOk) {
         ConsumerImplBasePtr consumer;
 
-        PULSAR_REGEX_NAMESPACE::regex pattern(regexPattern);
+        PULSAR_REGEX_NAMESPACE::regex pattern(TopicName::removeDomain(regexPattern));
 
         NamespaceTopicsPtr matchTopics =
             PatternMultiTopicsConsumerImpl::topicsPatternFilter(*topics, pattern);
 
-        consumer = std::make_shared<PatternMultiTopicsConsumerImpl>(
-            shared_from_this(), regexPattern, *matchTopics, subscriptionName, conf, lookupServicePtr_);
+        auto interceptors = std::make_shared<ConsumerInterceptors>(conf.getInterceptors());
+
+        consumer = std::make_shared<PatternMultiTopicsConsumerImpl>(shared_from_this(), regexPattern, mode,
+                                                                    *matchTopics, subscriptionName, conf,
+                                                                    lookupServicePtr_, interceptors);
 
         consumer->getConsumerCreatedFuture().addListener(
             std::bind(&ClientImpl::handleConsumerCreated, shared_from_this(), std::placeholders::_1,
@@ -338,8 +409,10 @@ void ClientImpl::subscribeAsync(const std::vector<std::string>& topics, const st
         topicNamePtr = TopicName::get(consumerTopicNameStream.str());
     }
 
+    auto interceptors = std::make_shared<ConsumerInterceptors>(conf.getInterceptors());
+
     ConsumerImplBasePtr consumer = std::make_shared<MultiTopicsConsumerImpl>(
-        shared_from_this(), topics, subscriptionName, topicNamePtr, conf, lookupServicePtr_);
+        shared_from_this(), topics, subscriptionName, topicNamePtr, conf, lookupServicePtr_, interceptors);
 
     consumer->getConsumerCreatedFuture().addListener(std::bind(&ClientImpl::handleConsumerCreated,
                                                                shared_from_this(), std::placeholders::_1,
@@ -383,20 +456,29 @@ void ClientImpl::handleSubscribe(const Result result, const LookupDataResultPtr
             conf.setConsumerName(generateRandomName());
         }
         ConsumerImplBasePtr consumer;
-        if (partitionMetadata->getPartitions() > 0) {
-            if (conf.getReceiverQueueSize() == 0) {
-                LOG_ERROR("Can't use partitioned topic if the queue size is 0.");
-                callback(ResultInvalidConfiguration, Consumer());
-                return;
+        auto interceptors = std::make_shared<ConsumerInterceptors>(conf.getInterceptors());
+
+        try {
+            if (partitionMetadata->getPartitions() > 0) {
+                if (conf.getReceiverQueueSize() == 0) {
+                    LOG_ERROR("Can't use partitioned topic if the queue size is 0.");
+                    callback(ResultInvalidConfiguration, Consumer());
+                    return;
+                }
+                consumer = std::make_shared<MultiTopicsConsumerImpl>(
+                    shared_from_this(), topicName, partitionMetadata->getPartitions(), subscriptionName, conf,
+                    lookupServicePtr_, interceptors);
+            } else {
+                auto consumerImpl = std::make_shared<ConsumerImpl>(shared_from_this(), topicName->toString(),
+                                                                   subscriptionName, conf,
+                                                                   topicName->isPersistent(), interceptors);
+                consumerImpl->setPartitionIndex(topicName->getPartitionIndex());
+                consumer = consumerImpl;
             }
-            consumer = std::make_shared<MultiTopicsConsumerImpl>(shared_from_this(), topicName,
-                                                                 partitionMetadata->getPartitions(),
-                                                                 subscriptionName, conf, lookupServicePtr_);
-        } else {
-            auto consumerImpl = std::make_shared<ConsumerImpl>(
-                shared_from_this(), topicName->toString(), subscriptionName, conf, topicName->isPersistent());
-            consumerImpl->setPartitionIndex(topicName->getPartitionIndex());
-            consumer = consumerImpl;
+        } catch (const std::runtime_error& e) {
+            LOG_ERROR("Failed to create consumer: " << e.what());
+            callback(ResultConnectError, {});
+            return;
         }
         consumer->getConsumerCreatedFuture().addListener(
             std::bind(&ClientImpl::handleConsumerCreated, shared_from_this(), std::placeholders::_1,
@@ -423,12 +505,19 @@ void ClientImpl::handleConsumerCreated(Result result, ConsumerImplBaseWeakPtr co
         }
         callback(result, Consumer(consumer));
     } else {
-        callback(result, {});
+        // In order to be compatible with the current broker error code confusion.
+        // https://github.com/apache/pulsar/blob/cd2aa550d0fe4e72b5ff88c4f6c1c2795b3ff2cd/pulsar-broker/src/main/java/org/apache/pulsar/broker/service/BrokerServiceException.java#L240-L241
+        if (result == ResultProducerBusy) {
+            LOG_ERROR("Failed to create consumer: SubscriptionName cannot be empty.");
+            callback(ResultInvalidConfiguration, {});
+        } else {
+            callback(result, {});
+        }
     }
 }
 
-Future<Result, ClientConnectionWeakPtr> ClientImpl::getConnection(const std::string& topic) {
-    Promise<Result, ClientConnectionWeakPtr> promise;
+Future<Result, ClientConnectionPtr> ClientImpl::getConnection(const std::string& topic, size_t key) {
+    Promise<Result, ClientConnectionPtr> promise;
 
     const auto topicNamePtr = TopicName::get(topic);
     if (!topicNamePtr) {
@@ -439,15 +528,20 @@ Future<Result, ClientConnectionWeakPtr> ClientImpl::getConnection(const std::str
 
     auto self = shared_from_this();
     lookupServicePtr_->getBroker(*topicNamePtr)
-        .addListener([this, self, promise](Result result, const LookupService::LookupResult& data) {
+        .addListener([this, self, promise, key](Result result, const LookupService::LookupResult& data) {
             if (result != ResultOk) {
                 promise.setFailed(result);
                 return;
             }
-            pool_.getConnectionAsync(data.logicalAddress, data.physicalAddress)
+            pool_.getConnectionAsync(data.logicalAddress, data.physicalAddress, key)
                 .addListener([promise](Result result, const ClientConnectionWeakPtr& weakCnx) {
                     if (result == ResultOk) {
-                        promise.setValue(weakCnx);
+                        auto cnx = weakCnx.lock();
+                        if (cnx) {
+                            promise.setValue(cnx);
+                        } else {
+                            promise.setFailed(ResultConnectError);
+                        }
                     } else {
                         promise.setFailed(result);
                     }
@@ -508,6 +602,7 @@ void ClientImpl::closeAsync(CloseCallback callback) {
     state_ = Closing;
 
     memoryLimitController_.close();
+    lookupServicePtr_->close();
 
     auto producers = producers_.move();
     auto consumers = consumers_.move();
@@ -640,10 +735,7 @@ uint64_t ClientImpl::newConsumerId() {
     return consumerIdGenerator_++;
 }
 
-uint64_t ClientImpl::newRequestId() {
-    Lock lock(mutex_);
-    return requestIdGenerator_++;
-}
+uint64_t ClientImpl::newRequestId() { return (*requestIdGenerator_)++; }
 
 uint64_t ClientImpl::getNumberOfProducers() {
     uint64_t numberOfAliveProducers = 0;
@@ -669,4 +761,13 @@ uint64_t ClientImpl::getNumberOfConsumers() {
 
 const ClientConfiguration& ClientImpl::getClientConfig() const { return clientConfiguration_; }
 
+std::string ClientImpl::getClientVersion(const ClientConfiguration& clientConfiguration) {
+    std::ostringstream oss;
+    oss << "Pulsar-CPP-v" << PULSAR_VERSION_STR;
+    if (!clientConfiguration.getDescription().empty()) {
+        oss << "-" << clientConfiguration.getDescription();
+    }
+    return oss.str();
+}
+
 } /* namespace pulsar */
